/* This file is part of mtbe. See mtbe.c for copyright notice. */


#include <sys/types.h>	/* Posix requires this for some of the others	*/
#include <stdio.h>	/* Ansi-C standard I/O		*/
#include <stdlib.h>	/* Ansi-C standard library	*/
#include <signal.h>	/* Posix signals		*/
#include <errno.h>	/* Posix error number		*/
#include <termios.h>
#include <unistd.h>	/* Unix standard functions	*/
#include <sys/times.h>	/* system clock			*/
#include <sys/wait.h>	/* wait for child process	*/

#include "mtbe.h"
#include "exec.h"
#include "errorcode.h"


/* Define DEBUG_MSG if you want debugging messages from Execute(). */
/* #define DEBUG_MSG */


/* Executes program "cmd" with argument "arg" in a subprocess in
** work directory "wdir", redirecting stdin and stdout to "ifp" resp. "ofp".
** Waits until the subprocess terminates, then returns.
** Returns the subprocess duration in clock ticks in "*elapsed".
** ("elapsed" may be NULL)
** Function value:
**	FATAL		fatal error; essentially any error other than a SIGHUP
**	OK		successful completion
**	TRYAGAIN	terminated by a SIGHUP
*/
int Execute(const char *cmd, const char *arg,
	    const char *wdir, FILE *ifp, FILE *ofp, clock_t *elapsed)
{
  /* Save current time */
  struct tms tbuf;
  clock_t const start = times(&tbuf);

  /* Create a new process */
  int const pid = fork();

  /* Make sure it worked */
  if (pid==-1)
  {
    PERROR("Can't fork");
    return FATAL;
  }
  else if (pid!=0)
  {
    int status;
    int retval = FATAL; /* initialize to worst case */

    /* In parent: wait until the child process terminates */
    waitpid(pid,&status,0);

    /* Return duration */
    if (elapsed)
      *elapsed = times(&tbuf)-start;

    /* If the subprocess exited normally, check it's return code.
    ** Any error in the subprocess (i.e. non-zero return code) is
    ** assumed to be fatal.
    */
    if (WIFEXITED(status))
    {
#ifdef DEBUG_MSG
      printf("Debug: WEXITSTATUS is %d\n", WEXITSTATUS(status));
      fflush(stdout);
#endif
#ifdef BUGGY_ZMODEM
      /* Ahhrgg! Some z-modem implementations give bogus exit codes. */
      switch (WEXITSTATUS(status)) {
      case 0:
      case 128:
	      retval = OK;
	      break;
      default:
	      retval = FATAL;
	      break;
      }
#else
      retval = WEXITSTATUS(status) ? FATAL : OK;
#endif
    }

    /* If the subprocess was terminated by an uncatched signal, then return
    ** TRYAGAIN if it was a SIGHUP.
    */
    if (WIFSIGNALED(status))
    {
#ifdef DEBUG_MSG
      printf("Debug: WTERMSIG is %d\n", WTERMSIG(status));
      fflush(stdout);
#endif
      retval = WTERMSIG(status) == SIGHUP ? TRYAGAIN : FATAL;
    }

    return retval;
  }

  /* Child now */

  /* We don't want to call exit handlers from the child, so we use _exit()
  ** here; this means we have to flush buffers explicitly.
  ** -rv
  */

#ifdef ENABLE_SUID
  if (setgid(getgid()) == -1 || setuid(getuid()) == -1)
  {
    PERROR("Cannot set uid/gid");
    fflush(stderr);
    _exit(EXIT_FAILURE);
  }
#endif

  /* Change to work directory */
  if (chdir(wdir)==-1)
  {
    PERROR("Cannot change working directory");
    fflush(stderr);
    _exit(EXIT_FAILURE);
  }

  /* a SIGHUP should kill the child */
  signal(SIGHUP, SIG_DFL);

  /* Clean up file buffers */
  fflush(ofp);
  tcflush(fileno(ifp),TCIFLUSH);
  tcflush(fileno(ofp),TCOFLUSH);

  /* Redirect stdin to ifp, stdout to ofp */
  dup2(fileno(ifp),STDIN_FILENO);
  fclose(ifp);
  dup2(fileno(ofp),STDOUT_FILENO);
  fclose(ofp);

  /* Now execute the command.
  ** We can only pass one arg this way, but this is
  ** enough for this application.
  */
  execlp(cmd,cmd,arg,(char*)0);

  /* "... and I'll kill you if you come back..."
  ** -- Meat Loaf (well, more or less)
  */
  fprintf(stderr,"Cannot exec %s: %s (%d)\n",
    cmd,strerror(errno),errno);
  fflush(stderr);
  _exit(EXIT_FAILURE);

  /* Calm down ambitious compilers */
  return FAIL;
}


