/* This file is part of mtbe. See mtbe.c for copyright notice. */


#include <ctype.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "errorcode.h"
#include "mtbe.h"	/* Countdown()	*/
#include "modem.h"


/* Define DEBUG_MSG if you want some debugging messages.
** You shouldn't use this for a "production" version.
*/
/* #define DEBUG_MSG */


/* For timeout jumps */
static jmp_buf Jmp;

/* SIGALRM handler for read timeout */
static void dosigalrm(int sig)
{
  longjmp(Jmp,1);
}

/* The use of ReadFromModem() is sometimes quite tricky: Suppose
   you're searching for a pattern and get a TRYAGAIN error.  In this
   case you must check if the pattern might have started at the end of
   the returned string.  If not, it's easy: simply repeat the call to
   ReadFromModem() with the same arguments.  But in the opposite case,
   it's plain wrong to search just for the remainder of the pattern.
   You have to read until the end of the line first (i.e. call
   ReadFromModem() with pat==NULL and NoEmpty==0) and check "by hand"
   if the pattern matches, possibly looping until you don't get
   TRYAGAIN anymore.  The best thing is to avoid a TRYAGAIN in the
   first place by choosing the size of the buffer str large enough, so
   that you only get a TRYAGAIN if someting is _really_ wrong.  Then
   you may as well treat it like a FAIL. */

/* ReadFromModem() reads from fp into str.
** If pat==NULL, reads until CR or LF. Trailing white space is cut off.
** If pat!=NULL, reads until "pat" found. Str returns only the last line read,
** with the found pattern at its end.
** Ignores empty lines if pat!=NULL or NoEmpty!=0.
**
** If captfp!=NULL, then every read char is written to it.
**
** No timeout is checked if timeout<=0.
** Please NOTE: a handler for SIGALRM is installed if timeout>0!
**
** Return:
**      FATAL		EOF or other error
**      TIMEOUT		timeout
**	OK		ok
**	TRYAGAIN	str length exceeded (strsize-1 char's read but still
**			no end-of-line)
*/
int ReadFromModem(FILE *fp, FILE **captfp, const int timeout,
		  const int NoEmpty, const char *pat,
		  char *str, const size_t strsize)
{
  char *p;
  size_t left;
  const char *pptr;
  int retval;
  long volatile oldalarm = 0;
  void (*volatile oldsig)(int) = SIG_IGN;

  if (strsize <= 0)
    return TRYAGAIN; /* catch this trivial error as soon as possible;
			strsize==1 gets caught by the loop logic. */

  if (setjmp(Jmp)!=0)
  {
#ifdef DEBUG_MSG
    puts("Timeout.");
#endif
    signal(SIGALRM,oldsig);
    alarm(oldalarm);
    return TIMEOUT;
  }

  if (timeout>0)
  {
    oldalarm = alarm(timeout);
    oldsig = signal(SIGALRM,dosigalrm);
  }

  /* Init a few things.
  ** We do it here and not in the declaration to avoid some compiler warnings.
  */
  p = str;
  left = strsize - 1;
  pptr = pat;
  retval = TRYAGAIN; /* initialize as "line too long" */

  while(left>0)
  {
    int const c = fgetc(fp);
    if (c==EOF)
    {
#ifdef DEBUG_MSG
      printf("Error reading from modem.\n");
#endif
      signal(SIGALRM,oldsig);
      alarm(oldalarm);
      return FATAL;
    }
    *p++ = c;
    left--;

    /* Restart alarm clock */
    if (timeout>0)
      alarm(timeout);

    putchar(c);
    fflush(stdout);
    if (captfp && *captfp &&
	(fputc(c, *captfp) == EOF
#ifdef FLUSH_LOG
	 || fflush(*captfp) == EOF
#endif
	))
      *captfp = NULL; /* we don't try a fclose() because most likely
			 this would result in another error; so we
			 simply inhibit further write operations */

    /* Match pattern */
    /* This isn't perfect but it should do. */
    if (pat!=NULL)
      if (c==*pptr)
      {
	pptr++;
	if (*pptr=='\0')
	{
	  retval = OK; /* match */
	  break;
        }
      }
      else
      {
	pptr = pat;
	if (c==*pat)
	  pptr++;
      }

    if (c=='\n' || c=='\r')	/* hey, end-of-line! */
    {
      /* ignore empty lines or lines that didn't match our pattern */
      if ((NoEmpty && *str==c) || pat!=NULL)
      {
        p = str;
	left = strsize - 1;
      }
      else
      {
	retval = OK; /* done */
	break;
      }
    }
  } /* end of reading loop; while(left>0) */
  *p = '\0';

  alarm(oldalarm);
  signal(SIGALRM,oldsig);

  /* Cut off trailing white space, but only if we've read an entire line.
  ** NOTE: If we've successfully matched a pattern, then the trailing white
  **       space is part of the pattern and should not be removed!
  */
  if (retval==OK && pat==NULL)
  {
    for (p=str+strlen(str)-1 ; p >= str && isspace(*p) ; p--)
      ; /* nothing */
    p[1] = '\0'; /* this is ok, really! :-) */
  }

  return retval;
}


/* Transmits str to modem at fp
** Return:
**	OK	ok,
**	FATAL 	error
*/
int SendToModem(FILE *fp, const char *str)
{
  if (fputs(str,fp)==EOF || fflush(fp)==EOF)
  {
#ifdef DEBUG_MSG
    printf("Error writing to modem.\n");
#endif
    return FATAL;
  }

  return OK;
}


/* EatEcho() reads one full line from the modem.
** If the length of this line exceeds the buffer length (given in strsize),
** only the last piece of it is returned.
** Return: see ReadFromModem(), except that TRYAGAIN is catched here...
*/
static int EatEcho(FILE *ifp, FILE **captfp, char *str, const int strsize)
{
  int rt, NoEmpty = 1;

  do {
    rt = ReadFromModem(ifp, captfp, 5, NoEmpty, NULL, str, strsize);
    NoEmpty = 0;
  } while (rt == TRYAGAIN);
  return rt;
}


/* InitModem sends the initialization strings to the modem
** Initstr: array of initialization strings
** ifp, ofp: in/out file pointers
** captfp: capture file
**
** Return:
**	OK	successful
**	FAIL	failure
*/
int InitModem(char *Initstr[], FILE *ifp, FILE *ofp, FILE **captfp)
{
  char str[1024+1];
  int i;

  /* "Wake up" the modem */
  if (SendToModem(ofp,"AT\r")!=0)
    return FAIL;

  for (i=0 ; ; i++)
  {
    /* this "7" is for compatibility with earlier versions; _must_ be >= 2 */
    if (i == 7 ||
	ReadFromModem(ifp, captfp, 5, 1, NULL, str, sizeof(str)) != OK)
      return FAIL;
    if (strcmp(str,"OK")==0)
      break;
  }

  /* Send all specified init commands */
  while (*Initstr != NULL)
  {
    if (SendToModem(ofp, *Initstr) != 0 || SendToModem(ofp, "\r") != 0)
      return FAIL;

    /* Ignore the echo, then get and check the answer */
    if (EatEcho(ifp, captfp, str, sizeof(str)) != OK ||
	ReadFromModem(ifp, captfp, 5, 1, NULL, str, sizeof(str)) != OK)
      return FAIL;
    if (strcmp(str,"OK")!=0)
      return FAIL;
    Initstr++;
  }
  return OK;
} /* InitModem */



/* Connect() tries to connect to the mailbox host. Retries if busy.
** Tries: number of allowed tries; returns number of remaining tries.
**	  Tries==NULL leads to unlimited retries.
**	  *Tries <= 0 leads to immetiate return with error code TRYAGAIN.
** Delay: time to sleep between two tries, in seconds.
** ifp, ofp: pointer to in/out streams connected to modem 
** Return:
**	FATAL		modem chat failed (e.g. no telephone line)
**	OK		ok
**	TRYAGAIN	busy (retries exceeded; doesn't happen if Tries==NULL)
**	FAIL		modem dialed, but failed (e.g. "NO CARRIER")
*/
int Connect(const char *PhoneNumber, int *Tries, const int Delay,
	    FILE *ifp, FILE *ofp, FILE **captfp)
{
  char str[1024+1];
  int do_countdown = 0;

  for(;;)
  {
    /* First check if max. number of tries exhausted;
    ** Tries==NULL means loop forever.
    */
    if (Tries)
      if (*Tries <= 0)
      {
#ifdef DEBUG_MSG
	puts("\nGiving up. (no more retries)");
#endif
	return TRYAGAIN;
      }
      else
        (*Tries)--; /* this must not be done in the "if" condition, or
		       else the returned value is wrong! */

    /* This doesn't look pretty, but calling Countdown() below (where
    ** do_countdown is set) has the drawback that the Tries check would have
    ** to happen there, too! (otherwise mtbe would wait before recognizing
    ** that no more retry is allowed)
    ** Now you can argue which ugliness would be better.
    */
    if (do_countdown)
    {
      do_countdown = 0;
      Countdown(Delay);
    }

    if (SendToModem(ofp, "ATD") != 0 || SendToModem(ofp, PhoneNumber) ||
	SendToModem(ofp, "\r") != 0)
      return FATAL;
    /* Ignore the echo, even for an extremely long PhoneNumber */
    if (EatEcho(ifp, captfp, str, sizeof(str)) != OK)
      return FATAL; /* no echo returned, really weird! */

    /* Wait for result, ignoring multiple RINGING messages (Zyxels do this) */
    do {
      if (ReadFromModem(ifp, captfp, 90, 1, NULL, str, sizeof(str)) != OK)
	return FATAL;
    } while (strcmp(str,"RINGING")==0);


    /* Check modem response: */

    /* Connected ok */
    if (strncmp(str,"CONNECT",7)==0 || strncmp(str,"CARRIER",7)==0)
      return OK;

    /* No telephone line, or more likely temporary line failure */
    if (strncmp(str,"NO DIAL",7)==0)
    {
      sleep(1); /* hmm, maybe the sleeping time should be larger... */
      continue;
    }

    /* Dial delayed */
    if (strncmp(str,"DELAY",5)==0 || strstr(str,"LOCK"))
    {
      sleep(5);
      continue;
    }

    /* Busy; we call Countdown() indirectly */
    if (strcmp(str,"BUSY")==0)
    {
      do_countdown = 1;
      continue;
    }

    /* Anything else */
#ifdef DEBUG_MSG
    printf("\nCannot connect.\n");
#endif
    return FAIL;
  }
} /* Connect() */

