/****************************************************************************
Please stand up and take off your hat. Listen raptly to the Copyright Notice.

This program was written by

	Wolfram Rsler		E-Mail: wr@bara.oche.de
	Augustastr. 44-46	Mausnet: @ AC2
	D-52070 Aachen		Voice: +49(0)241 534596
	Germany

Version 3.0 of this program was done by

        Robert Vogelgesang	E-Mail: rvogel@kuckuck.folz.de
	Dornenstr. 28		Mausnet: @ KL
	D-67657 Kaiserslautern
	Germany

The copyright remains with the authors.
This program is Freeware and may be used, copied and modified for any
non-commercial purpose as long as the copyright notices stay intact.
The program may be distributed without any restriction.
When modifying mtbe and distributing a modified version, please
change the version number (#define VERSION, see below), document
your changes in the source code and man page, and mail a context
diff of your changes to wr@bara.oche.de.

You may now sit down again.
*****************************************************************************/

/* MAUSTAUSCH BACKEND
** by Wolfram Rsler @ AC2 <wr@bara.oche.de>
** and Robert Vogelgesang @ KL
**
** This program is intended to be used as a Maustausch backend
** for O4X. It should compile cleanly on most Posix systems.
** No config file is needed, all parameters are passed on the
** command line. See the man page for details.
**
** The program requires rz and sz to be available.
** See the notes in the Makefile concerning some buggy implementations!
**
**
** CREDITS
**
** The intitial versions of the locking functions (LockPort and UnlockPort)
** were essentially taken from a "public domain" program by:
**		Karl Fox <karl@MorningStar.Com>
**		1760 Zollinger Road
**		Columbus, OH  43221, U.S.A.
** The -r option was introduced by Frank Baumgart @ PB3.
**
**
**
** For a HISTORY of mtbe see the file History.
**
** For a list of missing features see the file ToDo.
**
**
** KNOWN SECURITY RELATED PROBLEMS
**
** 1. If you want to set any SUID/SGID bits, be sure to compile mtbe with
**    ENABLE_SUID set.
** 2. The user may re-program the modem via "-z".
** -rv
**
**
** ATTENTION HACKERS
** Add your initials to the user name list in the History file and use them
** in comments to sign your changes!
*/


#include <sys/types.h>	/* Posix requires this for some of the others	*/
#include <stdio.h>	/* Ansi-C standard I/O		*/
#include <stdlib.h>	/* Ansi-C standard library	*/
#include <string.h>	/* Ansi-C string functions	*/
#include <ctype.h>	/* Ansi-C character types	*/
#include <time.h>	/* Ansi-C time functions	*/
#include <getopt.h>	/* argv option parser		*/
#include <errno.h>	/* Posix error number		*/
#include <fcntl.h>	/* Unix file control		*/
#include <unistd.h>	/* Unix standard functions	*/
#include <termios.h>	/* Posix terminal I/O stuff	*/
#include <sys/stat.h>	/* file status information	*/
#include <sys/times.h>	/* defines clock_t		*/

/* OK, forget about POSIX for a moment.  We'll use getpass(), for convenience.
** Is there a more portable way to get the getpass prototype into the program?
*/
extern char *getpass(const char *prompt);


#ifdef ENABLE_SUID
#include <limits.h>	/* Implementation limits	*/
#include "secure-log.h"
#endif


/* Define this if you suspect SIGHUP problems.		*/
/* #define DEBUG_SIGHUP */


#ifdef DEBUG_SIGHUP
#include <signal.h>	/* Posix signals		*/
int sighup_counter = 0;
#endif


#include "mtbe.h"
#include "term.h"
#include "modem.h"
#include "lock.h"
#include "exec.h"
#include "errorcode.h"


/* mtbe's version number */
#define VERSION "3.0"

/* some limits */
#define MAX_LOGIN_LENGTH	100	/* length of login names */
#define MAX_PWD_LENGTH		100	/* length of password */
#define MAX_BUF_LENGTH	       1024	/* length of file paths and
					   of strings read from modem */


/* structure for MultiTausch contexts */
typedef struct TauschContext
{
	char *Login;			/* Login name			*/
	char *pwdFile;			/* Password file name		*/
	char Passwd[MAX_PWD_LENGTH+1];	/* Password			*/
	char *Infile;			/* Infile name 			*/
	char *Outdir;			/* Outfile directory		*/
	struct TauschContext *next;	/* the next one in the list	*/
} TauschContext;


/* The name of this program (used by PERROR in mtbe.h) */
char *progname = "";

/* Capture file */
static FILE *captfp = NULL;


/* -------  End of declaration blurb. Start with actual code!  -------
** Sections:
**   1.  "Tausch" routines
**   2.  Context routines
**   3.  Other routines
**   4.  Main program
*/


/* -------  Section 1:  "Tausch" routines  ------- */


/* Do the first half of the Maustausch protocol (possibly in a "Multitausch"
** sequence).
** In a Maustausch, IsLast must not be 0 (zero).
** In a Multitausch, IsLast must be 0 (zero) for all but the last invocations.
**
** If DelPW!=0, then the password gets erased as soon as possible.
**
** Function value:
**	FATAL 		protocol error
**	OK		successful completion
**	TRYAGAIN	hangup
**	FAIL		(other) error during file transfer
*/
static int PutInfile(TauschContext *CtxPtr, const int IsLast, const int DelPW,
		     FILE *ifp, FILE *ofp)
{
  char str[MAX_BUF_LENGTH+1];
  clock_t ticks;		/* file transmission time	*/
  char *base;			/* basename			*/
  struct stat S;		/* for stat()'ing files		*/

  /* Maustausch protocol id */
  switch (ReadFromModem(ifp, &captfp, 60, 0, "eingetragener Benutzer? (J/N) ",
			str, sizeof(str)))
  {
    case FATAL: return TRYAGAIN;
    case OK: break;
    default: return FATAL;
  }
  if (SendToModem(ofp, IsLast ? "MausTausch\n" : "MultiTausch\n") != OK)
    return TRYAGAIN;

  /* Login name */
  switch (ReadFromModem(ifp, &captfp, 60, 0, "ame      : ", str, sizeof(str)))
  {
    case FATAL: return TRYAGAIN;
    case OK: break;
    default: return FATAL;
  }
  if (SendToModem(ofp, CtxPtr->Login) != 0 || SendToModem(ofp, "\r\n") != OK)
    return TRYAGAIN;

  /* Send password */
  switch (ReadFromModem(ifp, &captfp, 60, 0, "assword  : ", str, sizeof(str)))
  {
    case FATAL: return TRYAGAIN;
    case OK: break;
    default: return FATAL;
  }
  if (SendToModem(ofp, CtxPtr->Passwd) != OK)
    return TRYAGAIN;
  if (DelPW)
  {
    int i;
    /* Delete string immediately for safety */
    for(i=0;CtxPtr->Passwd[i];CtxPtr->Passwd[i++]='\0')
      ; /* nothing */
  }
  if (SendToModem(ofp, "\r\n") != OK)
    return TRYAGAIN;

  /* Send file transfer protocol id */
  if (SendToModem(ofp,"Z\r\n") != OK)
    return TRYAGAIN;

  switch (ReadFromModem(ifp, &captfp, 120, 0, "Protokoll startet:",
			str, sizeof(str)))
  {
    case FATAL: return TRYAGAIN;
    case OK: break;
    default: return FATAL;
  }

  printf("\n--- SENDING INFILE ---\n");
  fflush(stdout);	/* to avoid confusion with output of the subprocess */

  /* Put working directory into str; derive it from Infile; base is then the
  ** remainder of the Infile path
  */
  strcpy(str,CtxPtr->Infile);
  base = strrchr(str,'/');
  if (base==NULL)
  {
    base = CtxPtr->Infile;
    strcpy(str,".");
  }
  else
    *base++ = '\0';

  switch (Execute("sz", base, str, ifp, ofp, &ticks))
  {
    case TRYAGAIN: return TRYAGAIN;
    case OK: break;
    default: return FAIL;
  }

  /* Print cps rate; we need the first condition to avoid a division by zero */
  if (ticks > 0 && stat(CtxPtr->Infile, &S) == 0)
  {
    long   const bytes = (long)S.st_size;
    double const secs  = (double)ticks / CLK_TCK;
    double const cps   = (double)bytes / secs;
    char const fmt[] =
	    "\nInfile throughput: %ld bytes in %.1f seconds (%.1f cps)\n";
    printf(fmt, bytes, secs, cps);
    if (captfp && fprintf(captfp, fmt, bytes, secs, cps) < 0)
      captfp = NULL;
  }

  return OK;
} /* PutInfile */



/* Do the second half of the Maustausch protocol.
** Function value:
**	FATAL 		protocol error
**	OK		successful completion
**	TRYAGAIN	hangup
**	FAIL		(other) error during file transfer
*/
static int GetOutfile(TauschContext *CtxPtr, FILE *ifp, FILE *ofp)
{
  char str[MAX_BUF_LENGTH+1];	/* buffer for strings read from modem */
  clock_t ticks;		/* file transmission time	*/

  /* Some Maus BBS take a long time to get here; hence no timeout. */
  switch (ReadFromModem(ifp, &captfp, 0, 0, "Protokoll startet:",
			str, sizeof(str)))
  {
    case FATAL: return TRYAGAIN;
    case OK: break;
    default: return FATAL;
  }

  printf("\n--- RECEIVING OUTFILE ---\n");
  fflush(stdout);	/* to avoid confusion with output of the subprocess */

  switch (Execute("rz", NULL, CtxPtr->Outdir, ifp, ofp, &ticks))
  {
    case TRYAGAIN: return TRYAGAIN;
    case OK: break;
    default: return FAIL;
  }

  /* Print cps rate; this "if" is to avoid a division by zero */
  if (ticks > 0)
  {
    char *p;
    /* Construct outfile name first */
    if ((p=strrchr(CtxPtr->Infile,'.'))!=NULL)
    {
      char *tmp;
      size_t len;

      /* Length of "extension" */
      len = strlen(CtxPtr->Infile) - (p - CtxPtr->Infile);
      /* total length of outfile name, including '\0' */
      len += strlen(CtxPtr->Outdir) + 9;

      tmp = malloc(len);
      if (tmp)
      {
	struct stat S;

	sprintf(tmp, "%s/outfile%s", CtxPtr->Outdir, p);
	if (stat(tmp, &S) == 0)
	{
	  long   const bytes = (long)S.st_size;
	  double const secs  = (double)ticks / CLK_TCK;
	  double const cps   = (double)bytes / secs;
	  char const fmt[] =
	    "\nOutfile throughput: %ld bytes in %.1f seconds (%.1f cps)\n";
	  printf(fmt,bytes,secs,cps);
	  fflush(stdout);
	  if (captfp && fprintf(captfp,fmt,bytes,secs,cps) < 0)
	    captfp = NULL;
        }
	free(tmp);
      }
    }
  } /* print cps rate */

  return OK;
} /* GetOutfile */



/* -------  Section 2:  Context routines  ------- */


/* Allocates the storage for a new TauschContext.
** Returns NULL on failure.
*/
static TauschContext *NewContext(void)
{
  TauschContext *tmp;

  tmp = malloc(sizeof(struct TauschContext));
  if (!tmp)
  {
    fprintf(stderr, "%s: Couldn't get `Context' storage. (malloc() failed)\n",
	    progname);
    return NULL;
  }
  tmp->Login = NULL;
  tmp->pwdFile = NULL;
  tmp->Passwd[0] = '\0';
  tmp->Infile = "infile.txt";
  tmp->Outdir = ".";
  tmp->next = NULL;
  return tmp;
}

/* Checks if the context *Ctx has all we need: login name, password, infile
** strlength is the maximal value allowed for strlength(Infile)
*/
static int CheckContext(TauschContext *Ctx, const int strlength)
{
  /* Make sure we have the login name */
  if (Ctx->Login==NULL)
  {
    fprintf(stderr, "%s: Need login name (%s -h for help)\n",
	    progname, progname);
    return FAIL;
  }

  /* check length of infile path  -rv */
  if (strlen(Ctx->Infile) > strlength)
  {
    fprintf(stderr, "%s: Infile name too long; maximum %d chars.\n",
	    progname, strlength);
    return FAIL;
  }
  /* Should we stat() the infile here, to detect some really silly command
     line errors?   -rv */

  /* Now get the password */
  if (Ctx->pwdFile)
  {
    /* Load password from a file */
    FILE *const fp = fopen(Ctx->pwdFile, "r");
    char *p;

    if (fp==NULL)
    {
      fprintf(stderr, "%s: Can't open password file `%s'\n",
	      progname, Ctx->pwdFile);
      return FAIL;
    }

#ifdef ENABLE_SUID
    {
      /* Ok, we've opened the file. Now check if the user is allowed to
      ** read from the file; if not, complain and exit. We do it this
      ** way to avoid possible races.
      ** -rv
      */

      struct stat sb;

      fstat(fileno(fp), &sb);
      if ( !(sb.st_mode & S_IROTH) &&
	  (sb.st_uid != getuid() || !(sb.st_mode & S_IRUSR)) )
      {
        /* we have to check all groups the user's in */
	int size, g;
	gid_t glist[NGROUPS_MAX];
	size = getgroups(NGROUPS_MAX, glist);
	for (g=0; g<size && (glist[g]!=sb.st_gid) ; g++) ;
	if (g==size || !(sb.st_mode & S_IRGRP))
	{
	  fprintf(stderr, "%s: Can't read password file `%s'\n",
		  progname, Ctx->pwdFile);
	  fclose(fp);
	  return FAIL;
	}
      }
    }
#endif

    if (fgets(Ctx->Passwd, sizeof(Ctx->Passwd), fp) == NULL)
    {
      fclose(fp);
      fprintf(stderr, "%s: password file `%s' is empty\n",
	      progname, Ctx->pwdFile);
      return FAIL;
    }
    fclose(fp);

    /* Strip off trailing newline */
    if ((p=strrchr(Ctx->Passwd,'\n'))!=NULL)
      *p = '\0';
  }
  else if (Ctx->Passwd[0]=='\0')
  {
    int i;
    char *Pwd;
    char prompt[MAX_LOGIN_LENGTH+21+1];
    sprintf(prompt, "Enter password for %.*s: ", MAX_LOGIN_LENGTH,
	    Ctx->Login);
    Pwd = getpass(prompt);
    for (i=0 ; Pwd[i] && i<MAX_PWD_LENGTH ; i++)
      Ctx->Passwd[i]=Pwd[i], Pwd[i]='\0';
    Ctx->Passwd[i] = '\0';
  }

  /* Again: do we finally have a password? */
  if (Ctx->Passwd[0]=='\0')
  {
    fprintf(stderr, "%s: Need password for %s (%s -h for help)\n",
	    progname, Ctx->Login,progname);
    return FAIL;
  }
  return OK;
} /* CheckContext */



/* -------  Section 3:  Other routines  ------- */


static int IgnoreError(void)
{
  char ans[10];
  char *p;

  printf("\nFile transfer returned error. Continue anyway? (y/N) ");
  fflush(stdout);
  p = fgets(ans, 10, stdin);
  if (!p)
    return 0;
  while (isspace(*p))
    p++;
  return (*p == 'Y' || *p == 'y');
}


void Countdown(const int Delay)
{
  time_t const end = time(NULL) + Delay;

  printf("\n");
  for(;;)
  {
    time_t const now = time(NULL);

    /* The whole loop seems unneccessary complicated, but as it is, it's
    ** main advantage is that it prints the message for end-now>=0, but
    ** waits only if end-now>0. We have to catch now>end first, because
    ** the user may have given us something stupid like Delay==-1. But the
    ** clock ticks may have increased between the two calls to time(), too.
    */
    if (now>end)
      return;

    printf("\rWaiting... %-5d",(int)(end-now));
    fflush(stdout);
    if (now==end)
      return;
    sleep(1);
  }
} /* Countdown() */


/* Print usage summary and exit */
static void Usage(const char *pname)
{
  fprintf(stderr,
    "Maustausch Backend\n"
    "Usage: %s { options }\n"
    "Options are:\n"
    "\t-c x\t  Capture all communication into file `x'\n"
    "\t-t x\t  Use modem device `x'\n"
    "\t-b x\t  Use comm speed `x' bps\n"
    "\t-z x\t  Use modem init command `x' (multiple use allowed)\n"
    "\t-n x\t* Call telephone number `x'\n"
    "\t-d x\t  When busy, wait `x' seconds before retry\n"
    "\t-r x\t  Make at most `x' connection attempts (default: infinity)\n"
    "\t-R x\t  Allow connection attempts after up to `x' errors (default: 0)\n"
    "\t-l x\t* Log in with user name `x'\n"
    "\t-p x\t  Log in with password `x', enter interactively if empty\n"
    "\t-p @x\t  Load password from file `x'\n"
    "\t-i x\t  Send infile `x'\n"
    "\t-o x\t  Receive outfile in directory `x'\n"
    "\t-M\t  Separates MultiTausch user arguments\n"
    #ifndef LOCKMODE
    "\t-A\t  Use ascii lock file\n"
    "\t-L\t  Use binary lock file\n"
    #endif
    #ifndef ENABLE_SUID
    "\t-D x\t  Create lock file in directory `x'\n"
    #endif
    "\t-v\t  Print version number and exit\n"
    "\t-h\t  Print usage help and exit\n"
    "\t\t* = obligatory\n"
    "Default:\n"
    "\t%s -t /dev/modem -b 19200 -z ATZ -i infile.zip -o . -d 60\n"
    ,pname,pname);

  exit(EXIT_FAILURE);
}


/* Wrapper around strtol() with the same interface as atoi(),
** but it calls Usage() in case of an error.
*/
static long int Atoi(const char *str)
{
  char *end;
  long int val = strtol(str, &end, 0);

  if (!end || *end != '\0' || str == end)
    Usage(progname);

  return val;
}


/* "19200" -> B19200 etc.
** If error, exit()s.
*/
static int strtospeed(const char *str)
{
  int const bps = atoi(str);
  switch(bps)
  {
    case 1200:	return B1200;
    case 2400:	return B2400;
    case 4800:	return B4800;
    case 9600:	return B9600;
    case 19200:	return B19200;
    case 38400:	return B38400;
  }

  printf("Use drums for %s bps, joker\n",str);
  exit(EXIT_FAILURE);
}


#ifdef DEBUG_SIGHUP
static void count_sighup(int sig)
{
  sighup_counter++;
  signal(SIGHUP, count_sighup);
}
#endif



/* -------  Section 4:  Main program  ------- */


int main(int argc,char **argv)
{
  FILE *ifp, *ofp;		/* in/out file pointers for modem	*/
  int ifd, ofd;			/* file descriptor equivalents for ifp, ofp */
  int c;			/* for option parsing		*/
  char str[MAX_BUF_LENGTH+1];   /* read buffer (modem)		*/
  char *Modem = "/dev/modem";	/* modem device			*/
  char *Init[32+1] = { "ATZ", NULL };	/* init strings		*/
  char **InitPtr = Init;	/* for operations on Init	*/
  char *Tel = NULL;		/* string for dialing		*/
  char *LockDir = LOCK_DIR;	/* path to lock file directory	*/
  char *Capture = NULL;		/* Name of capture file		*/
  int Delay = 60;		/* Dial delay in seconds	*/
  int Tries = -1;		/* Connect retries, -1 = infinity */
  int ReTries = 0;		/* Retries after error, 0 = none */
  int reset_dial_cnt = 1;	/* reset dial counter for each retry? */
  int dial_cnt, retry_cnt;	/* counters			*/
  int retval;			/* for check of return values	*/
  int Speed = B19200;		/* terminal line speed		*/
  enum { none, lockascii, lockbinary } lockmode = none;
  time_t Online, Start;		/* Online and Connect time	*/
  TauschContext *First, *Last, *CtxPtr; /* list of contexts	*/
  extern int optind;
  extern char *optarg;


  /* Put program name into global variable so it can be
  ** be used by the PERROR macro
  */
  progname = argv[0];
 
  /* Set forced lock mode if there is one  -rv */
#ifdef LOCKMODE
  lockmode = LOCKMODE;
#endif

  /* Init the context list */
  First = Last = CtxPtr = NewContext();
  if (!First)
    return EXIT_FAILURE;

  /* Scan command line arguments */
  while((c=getopt(argc,argv,"Ab:c:D:d:hi:o:R:r:t:Mn:Ll:p:z:v?"))!=EOF)
    switch(c)
    {
      #ifndef LOCKMODE
      case 'A': lockmode = lockascii;break;
      case 'L': lockmode = lockbinary;break;
      #endif
      case 'b': Speed = strtospeed(optarg);break;
      case 'c': Capture = optarg;break;
      #ifndef ENABLE_SUID
      case 'D': LockDir = optarg;break;
      #endif
      case 'd': Delay = Atoi(optarg);break;
      case 'R':
	ReTries = Atoi(optarg);
	if (ReTries < 0)
	{
	  ReTries = -ReTries;
	  reset_dial_cnt = 0;
	}
	break;
      case 'r': Tries = Atoi(optarg);break;
      case 'i': CtxPtr->Infile = optarg;break;
      case 'o': CtxPtr->Outdir = optarg;break;
      case 't': Modem = optarg;break;
      case 'n': Tel = optarg;break;
      case 'l': CtxPtr->Login = optarg;break;
      case 'z':
        if (InitPtr-Init<sizeof(Init)/sizeof(*Init)-1)
	{
	  *InitPtr++=optarg;
	  InitPtr[1]=NULL;
	}
	else
	{
	  fprintf(stderr, "%s: Too many `-z' options!\n", progname);
	  return EXIT_FAILURE;
	}
	break;

      case 'p':
	if (optarg[0] == '@')
	{
	  CtxPtr->pwdFile = optarg + 1;
	  CtxPtr->Passwd[0] = '\0';
        }
	else
	{
	  /* copy password and delete it in argv for safety
	  ** (don't want ps(1) to reveal the password)
	  ** It should be safe enough to erase only the significant characters.
	  ** If you do not agree, then DON'T put the password into the command
	  ** line. Period. (use a file or give it interactively)
	  */
	  char *i,*o;
	
	  for(i=optarg, o=CtxPtr->Passwd ;
	      *i!='\0' && o - CtxPtr->Passwd < MAX_PWD_LENGTH ;
	      i++, o++)
	    *o=*i, *i='\0';
	  *o = '\0';
	  CtxPtr->pwdFile = NULL;
        }
	break;

      case 'M':
	/* Get a new Multitausch context and tack it to the end of the list */
	CtxPtr = NewContext();
	if (!CtxPtr)
	  return EXIT_FAILURE;
	Last = Last->next = CtxPtr;
	break;

      /* Print version number */
      case 'v':
        printf(
	  "Maustausch Backend version " VERSION ", compiled " __DATE__"\n"
	  "Copyright 1995, 1996 Wolfram Roesler @ AC2 <wr@bara.oche.de>\n"
	  "and Robert Vogelgesang @ KL <rvogel@kuckuck.folz.de>\n"
	  "May be copied and modified for any non-commercial purpose.\n");
        exit(EXIT_SUCCESS);

      default : Usage(progname);
    }

  /* Allow no further args */
  if (argc==1 || argv[optind]!=NULL)
    Usage(progname);

  /* Make sure we have the phone number */
  if (Tel==NULL)
  {
    fprintf(stderr,"%s: Need telephone number (%s -h for help)\n",
      progname,progname);
    return EXIT_FAILURE;
  }

  /* Check all contexts; we _need_ login name, password, infile. */
  for(CtxPtr=First; CtxPtr; CtxPtr = CtxPtr->next)
  {
    if (CheckContext(CtxPtr, MAX_BUF_LENGTH) != OK)
      return EXIT_FAILURE;
  }

  /* Optionally prefix the modem port with device directory  -rv */
  if (*Modem != '/' && *Modem != '.')
  {
    char *const t = malloc(strlen(DEV_DIR)+strlen(Modem)+1);
    if (t)
    {
      strcpy(t, DEV_DIR);
      strcat(t, Modem);
      Modem = t;
      /* free()d by exit */
    }
    else
    {
      fprintf(stderr, "%s: Can't allocate memory for string operations!\n",
	progname);
      return EXIT_FAILURE;
    }
  }

  /* Lock modem port if so desired */
  if (lockmode!=none)
  {
    char *LockBase;

    /* get basename of Port */
    LockBase = strrchr(Modem,'/');
    LockBase = LockBase ? LockBase+1 : Modem;
    if (LockPort(lockmode==lockbinary, LockDir, LOCK_PREFIX, LockBase))
    {
      fprintf(stderr,"%s is locked, mtbe terminated\n",Modem);
      return EXIT_FAILURE;
    }

    /* Make sure port gets unlocked on exit */
    atexit(UnlockPort);
  }


  /* Open capture file if needed */
  if (Capture!=NULL)
  {
#ifdef ENABLE_SUID
    /* for _real_ security, spawn a log daemon that drops its privileges;
    ** this is the only way to go if we want to use only Posix.
    ** The "NULL" tells this function that we don't want to install our
    ** own SIGCHLD handler.  -rv
    */
    captfp = SpawnLogDaemon(Capture, NULL);
#else
    captfp = fopen(Capture,"a");
    if (!captfp)
      fprintf(stderr, "%s: Warning: can't open capture file `%s'.\n",
	progname, Capture);
#endif
    /* set close-on-exec */
    if (captfp)
    {
      int flags = fcntl(fileno(captfp), F_GETFD);
      flags |= FD_CLOEXEC;
      (void) fcntl(fileno(captfp), F_SETFD, flags);
    }
  }


  /* loop over all allowed connect (re)tries */

  dial_cnt = Tries;	/* needed only if reset_dial_cnt==0 */
  for (retval = TRYAGAIN, retry_cnt = 0 ;
       retval == TRYAGAIN && retry_cnt <= ReTries ;
       retry_cnt++)
  {

    /* Open I/O files for modem */
    /* IMPORTANT: (ENABLE_SUID security!)
    **	o  use "O_RDWR" instead of "O_WRONLY" to avoid truncation if applied
    **	   to ordinary files (user error or malicious attack)
    **	o  use "dup()" instead of a second "open()" to avoid a race
    ** Not related to security:
    **	o  use "O_NOCTTY" to avoid hassles when called via cron
    **	o  use "O_NONBLOCK" to avoid a delay on open
    ** -rv
    */
    if ( (ofd = open(Modem, O_RDWR|O_NOCTTY|O_NONBLOCK)) < 0 )
    {
      fprintf(stderr, "%s: Can't open %s: %s (%d)\n",progname,Modem,
	      strerror(errno),errno);
      return EXIT_FAILURE;
    }

    /* Cancel the O_NONBLOCK flag. */
    (void) fcntl(ofd, F_SETFL, fcntl(ofd, F_GETFL) & ~O_NONBLOCK);

    if ( (ifd = dup(ofd)) < 0 ||
	 !(ofp = fdopen(ofd, "w")) || !(ifp = fdopen(ifd, "r")) )
    {
      fprintf(stderr, "%s: Can't duplicate fd's for %s: %s (%d)\n",
	      progname, Modem, strerror(errno), errno);
      return EXIT_FAILURE;
    }

    /* Make sure it is a char dev. This is just for beauty; without this
    ** section the call to Configure() would detect the same error. -rv
    */
    {
      struct stat mstat;
      if (fstat(ofd, &mstat) == -1 || (!S_ISCHR(mstat.st_mode)))
      {
	fprintf(stderr, "%s: %s is not a character device.\n",
		progname, Modem);
	return EXIT_FAILURE;
      }
    }

    /* configure the modem port */
    /* Well, ifp and ofp both point to the same device. And all of
    ** tc[gs]etattr() and cfset[io]speed() work with devices instead of file
    ** descriptors. Hence it's sufficient to "Configure" only one of ifp/ofp.
    */
    if (Configure(ifp, Speed) != 0)
      return EXIT_FAILURE;


#ifdef DEBUG_SIGHUP
    /* Install a SIGHUP handler; this is for debugging only */
    signal(SIGHUP, count_sighup);
#endif


    /* Read anything the modem might want to tell us. This is important
    ** in the case of a retry: Some modems won't re-dial without this!
    ** The only possible limit for this purpose is a timeout; 1 second should
    ** be enough. (the 1 for NoEmpty may as well be a 0, doesn't matter here)
    ** (the ">" instead of a "!=" is to avoid an endless loop for FATAL)
    ** For added reliability, we do this even for the first try.
    */
    while (ReadFromModem(ifp, &captfp, 1, 1, NULL, str, sizeof(str)) > TIMEOUT)
      ; /* nothing */

    /* We should wait here before a retry... */
    if (retry_cnt > 0)
    {
      Countdown(Delay);
      printf("\n");
    }


    /* At this point we're starting to talk to the modem.
    ** The modem code does write to stdout, and because we want to preserve the
    ** order of messages as far as possible, all subsequent _informational_
    ** messages go to stdout, too. Only real error messages go to stderr.
    */


    /* Send all initialization strings to the modem */
    if (InitModem(Init, ifp, ofp, &captfp) != OK)
    {
      fprintf(stderr, "\n%s: Modem initialization failed.\n", progname);
      return EXIT_FAILURE;
    }

    /* Get start time */
    Start = time(NULL);

    /* Dial */
    if (reset_dial_cnt)
      dial_cnt = Tries;
    switch (Connect(Tel, (dial_cnt<0 ? NULL : &dial_cnt), Delay,
		    ifp, ofp, &captfp))
    {
      case OK: break;
      case FAIL: continue; /* retval is still == TRYAGAIN (retry, option -R) */
      case FATAL:
	fprintf(stderr, "\n%s: Huh, something's severely wrong.\n", progname);
	fprintf(stderr, "%s: (Possibly a hardware problem?)\n", progname);
	/* FALL THROUGH */
      default:
	/* typically: no more retries allowed; option -r */
	if (Tries >= 0)
	{
	  if (dial_cnt == 0)
	    fprintf(stderr, "\n%s: Retry limit (-r) reached.\n", progname);
	  dial_cnt = Tries - dial_cnt;
	  fprintf(stderr, "%s: (stopped after %d %s.)\n", progname,
		  dial_cnt, (dial_cnt==1 ? "try" : "tries"));
	}
        return EXIT_FAILURE;
    }

    /* Remember online time */
    Online = time(NULL);

    printf("\n--- ONLINE ---\n");

    /* Do a poll for all contexts.
    ** We remove the current context from the list if the Tausch was
    ** successful. Hence the somewhat "strange" while statement.
    ** (yes, an assignment!)
    */
    while ((CtxPtr = First) != NULL)
    {
      retval = PutInfile(CtxPtr, CtxPtr->next==NULL, ReTries==0, ifp, ofp);
      if (retval == OK || (retval == FAIL && IgnoreError()))
        retval = GetOutfile(CtxPtr, ifp, ofp);
      else
        break;

      /* if we have a FAIL here, then it's coming from GetOutfile() */
      if (retval == FAIL && IgnoreError())
        retval = OK;

      if (retval != OK)
        break;
      /* This was successful. Remove the context from the list.
      ** We can safely assume that we're at the start of the list.
      */
      First = First->next;
      free(CtxPtr);
    } /* do a poll for all contexts */

    /* If we have a OK here, then we're done. */
    if (retval == OK)
    {
      /* Wait until the connection is closed; we expect the modem to hang up,
      ** so we set CLOCAL first or else ReadFromModem() won't get what we need.
      */
      /* no error handling here, we'll close down anyway */
      (void) SetCLOCAL(ifp);
      (void) ReadFromModem(ifp, &captfp, 60, 0, "NO CARRIER",
			   str, sizeof(str));
    }
    /* Close tty files. We _need_ this even in the case of a pending retry! */
    RestoreTermFlags();
    fclose(ifp);
    fclose(ofp);
    close(ofd);
    close(ifd);

    /* Print termination message and duration to screen and capture file. */
    {
      time_t const Now = time(NULL);
      int const T_Secs = (int)(Now-Start);
      int const O_Secs = (int)(Now-Online);
      const char fmt[] =
      "\n%s: Maustausch %s. (total time: %d' %02d\" online: %d' %02d\")\n";
      char *t;

      switch (retval)
      {
        case OK:
	  t = "complete";
	  break;
        case TRYAGAIN:
	  t = "incomplete, try again";
	  break;
        default:
	  t = "failed";
	  break;
      }
      printf(fmt, progname, t, T_Secs/60, T_Secs%60, O_Secs/60, O_Secs%60);
      if (captfp!=NULL && 
	  fprintf(captfp, fmt, progname, t,
		  T_Secs/60, T_Secs%60, O_Secs/60, O_Secs%60) < 0)
	captfp = NULL;
    }

  } /* retry-after-error loop */

  if (retval == TRYAGAIN)
  {
    const char fmt[] = "%s: Retry limit (-R) reached, giving up.\n";

    printf(fmt, progname);
    if (captfp!=NULL)
      fprintf(captfp, fmt, progname);	/* errors ignored; it's the last write
					   to the capture file anyway */
  }

  /* "And I think somebody somewhere must be tolling a bell ..." */
  putchar('\a');
  fflush(stdout);
  /* Free all remaining contexts and print a message for each of it. */
  while ((CtxPtr = First) != NULL)
  {
    fprintf(stderr, "%s: Maustausch failed for user `%s'.\n", progname,
	    CtxPtr->Login);
    First = First->next;
    free(CtxPtr);
  }

#ifdef DEBUG_SIGHUP
  printf("Debug: We've got %d SIGHUP%s.\n", sighup_counter,
	 (sighup_counter==1 ? "" : "s"));
#endif

  /* "... and the last thing I see is ..." */
  return (retval == OK? EXIT_SUCCESS : EXIT_FAILURE);
}

