/* This file is part of mtbe and subject to the following copyright:
 * Copyright 1996 Robert Vogelgesang
 * You may freely distribute, use and modify this file for any
 * purpose as long as you adhere to the following two conditions:
 *   1. The copyright notice (including this list of conditions) remains
 *	unchanged.
 *   2. If you use this code (modified or not) in any other context than mtbe,
 *	you must give credit to the copyright holder in the documentation or
 *	the source code of your program. That part of the documentation or the
 *	source code (the one with the credit note) must be supplied free of
 *	additional charge along with any compiled versions of your program.
 *	(i.e. you may not pretend you've written this code yourself)
 */


#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>

#include "mtbe.h"	/* defines PERROR */
#include "secure-log.h"


static pid_t daemonpid;

/* SIGCHLD handler for log daemon */
static void dosigchld(int sig)
{
	/* This is to avoid interference with other possibly dead child
	   processes. */
	if (waitpid(daemonpid, NULL, WNOHANG) != daemonpid)
		signal(SIGCHLD, dosigchld); /* re-register */
	return;
}


/* Spawn a log daemon.
 * Returns a pointer to the output stream on success, NULL on failure.
 * If the pointer "pidreturn" equals NULL, a private SIGCHLD handler is
 * installed; otherwise the pid of the created child process is returned
 * on success.
 * The simple SIGCHLD handler above can only cope with one single daemon
 * process. If you need more, you have to install your own handler. In this
 * case, you must use the "pidreturn" mechanism to avoid the installation
 * of the above SIGCHLD handler.
 * SIGPIPE gets blocked in any case. (If the daemon dies for some reason,
 * then the program should be able to continue.)
 */
FILE *SpawnLogDaemon(const char *logfilepath, int *pidreturn)
{
	int pipe_fd[2];

	/* Block SIGPIPE before doing anything else; this is to avoid a race
	 * condition if the daemon dies immediately.
	 */
	signal(SIGPIPE, SIG_IGN);
	
	if (pipe(pipe_fd) < 0 || (daemonpid = fork()) < 0) {
		PERROR("Can't create log daemon");
		return NULL;
	} else {
		if (daemonpid) {
			/* parent */
			close(pipe_fd[0]);
			if (pidreturn)
				*pidreturn = daemonpid;
			else
				signal(SIGCHLD, dosigchld);

			return fdopen(pipe_fd[1], "w");
		} else {
			/* child (=daemon) */
			FILE *infp, *outfp;
			int c = 1;	/* exit code in case of an error */

			/* Unblock SIGPIPE; this might be useful if the
			 * file is actually a named FIFO.
			 */
			signal(SIGPIPE, SIG_DFL);

			close(pipe_fd[1]);
			infp = fdopen(pipe_fd[0], "r");

			if (setgid(getgid()) == -1 || setuid(getuid()) == -1) {
				PERROR("Cannot set uid/gid");
				goto exit_daemon;
			}
			outfp = fopen(logfilepath, "a");
			if (outfp==NULL) {
				PERROR("Can't open log file");
				goto exit_daemon;
			}
			while ((c = fgetc(infp)) != EOF) {
				fputc(c, outfp);
#ifdef FLUSH_LOG
				fflush(outfp);
#endif
			}
			fclose(outfp);
			c = 0;	/* "good" exit code */
		exit_daemon:
			fclose(infp);
			fflush(stdout);
			fflush(stderr);
			_exit(c);
		}
	}
}

