/* This file is part of mtbe. See mtbe.c for copyright notice. */

#include <stdio.h>	/* Ansi-C standard I/O		*/
#include <stdlib.h>	/* Ansi-C standard library	*/
#include <termios.h>	/* Posix terminal I/O stuff	*/


#include "errorcode.h"
#include "mtbe.h"	/* macro PERROR()	*/
#include "term.h"


/* Struct and fd for remembering and restoring the termios flags */
static struct termios tdata_orig;
static int t_fd = -1;


/* Restore the termios flags */
void RestoreTermFlags(void)
{
  if (t_fd == -1)
    return;

  (void) tcsetattr(t_fd, TCSANOW, &tdata_orig);
  t_fd = -1;
}



/* Configure fp for modem operation and set to Speed
** Return: 0 ok, else error
*/
int Configure(FILE *fp,int Speed)
{
  struct termios tdata;
  int const fd = fileno(fp);

  /* Make sure all data gets transmitted */
  tcdrain(fd);

  /* If not already done, get and save attributes */
  if (t_fd == -1)
  {
    t_fd = fd;
    if (tcgetattr(fd,&tdata_orig)!=0)
    {
      PERROR("Error in tcgetattr");
      return FATAL;
    }

    /* Make sure flags get restored on exit */
    atexit(RestoreTermFlags);
  }

  tdata = tdata_orig;

  /* Set line attributes */
  tdata.c_iflag |= IGNBRK | IGNPAR;
  tdata.c_iflag &= ~(ISTRIP | ICRNL | INLCR | IGNCR | IXON | IXOFF);
  tdata.c_oflag &= ~OPOST;
  /* According to POSIX we have to clear CSIZE before setting CS8.
     Under Linux, it's no difference because of CSIZE == CS8. */
  tdata.c_cflag &= ~(CLOCAL | CSTOPB | PARENB | CSIZE);
  tdata.c_cflag |= CS8 | CREAD | HUPCL;
  tdata.c_lflag &= ~(ISIG | ICANON | TOSTOP | ECHO | ECHOE | ECHOK
		     | ECHONL | NOFLSH );
  tdata.c_cc[VMIN]  = 1 ; 
  tdata.c_cc[VTIME] = 0 ; 

  /* Set input and output line speed */
  if (cfsetispeed(&tdata,Speed)==-1 || cfsetospeed(&tdata,Speed)==-1)
  {
    PERROR("Error setting line speed");
    return FATAL;
  }

  /* Get rid of line noise */
  tcflush(fd,TCIOFLUSH);

  /* Set attributes */
  if (tcsetattr(fd,TCSANOW,&tdata)!=0)
  {
    PERROR("Error in tcsetattr");
    return FATAL;
  }

  return OK;
}


/* SetCLOCAL
** Sets the CLOCAL line attribute
** Return: 0 ok, else error
*/
int SetCLOCAL(FILE *fp)
{
  struct termios tdata;
  int const fd = fileno(fp);

  if (tcgetattr(fd,&tdata)!=0)
  {
    PERROR("Error in tcgetattr");
    return FATAL;
  }

  tdata.c_cflag |= CLOCAL;

  if (tcsetattr(fd,TCSANOW,&tdata)!=0)
  {
    PERROR("Error in tcsetattr");
    return FATAL;
  }

  return OK;
}

/* ClearCLOCAL
** Clears the CLOCAL line attribute
** Return: 0 ok, else error
*/
int ClearCLOCAL(FILE *fp)
{
  struct termios tdata;
  int const fd = fileno(fp);

  if (tcgetattr(fd,&tdata)!=0)
  {
    PERROR("Error in tcgetattr");
    return FATAL;
  }

  tdata.c_cflag &= ~CLOCAL;

  if (tcsetattr(fd,TCSANOW,&tdata)!=0)
  {
    PERROR("Error in tcsetattr");
    return FATAL;
  }

  return OK;
}

